"""UB matrix calculations.

.. currentmodule:: diffcalc.ub

Routines for defining crystal lattice and UB matrix calculation from reference
reflections and crystal reference orientations.

.. autosummary::
   :toctree: generated/

   calc - UB matrix calculation.
   crystal - Crystal lattice data object.
   fitting - Routines for refining crystal lattice parameters and U matrix.
   reference - Reference reflection and crystal orientations data object.

"""
